# -*- coding: utf-8 -*-
import numpy as np
from numba import njit, prange, prange

__cache = True


@njit(nogil=True, cache=__cache)
def show_vector(dcm: np.ndarray, arr: np.ndarray):
    return dcm @ arr


@njit(nogil=True, parallel=True, cache=__cache)
def show_vectors(dcm: np.ndarray, arr: np.ndarray):
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm @ arr[i, :]
    return res

@njit(nogil=True, parallel=True, cache=__cache)
def show_vectors(dcm: np.ndarray, arr: np.ndarray):
    res = np.zeros_like(arr)
    for i in prange(arr.shape[0]):
        res[i] = dcm @ arr[i, :]
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numba import jit


@jit(nopython=True, nogil=True, cache=True)
def backsub_fr(A: np.ndarray, B: np.ndarray, presc_bool: np.ndarray,
               presc_val: np.ndarray, eqpath: np.ndarray,
               glob_to_front: np.ndarray, glob_to_width: np.ndarray):
    nEQ, maxwidth = A.shape
    nEQ, nRHS = B.shape
    X = np.zeros((nEQ, nRHS), dtype=np.float64)
    R = np.zeros((nEQ, nRHS), dtype=np.float64)
    resid = np.zeros(nRHS, dtype=np.float64)
    X_fr = np.zeros((maxwidth, nRHS), dtype=np.float64)
    for iEQ in range(nEQ - 1, -1, -1):
        gind = eqpath[iEQ]
        ifront = glob_to_front[gind]
        frwidth = glob_to_width[gind]
        pivot = A[gind, ifront]
        A[gind, ifront] = 0.0
        resid[:] = B[gind]
        for jfront in range(frwidth):
            resid -= A[gind, jfront] * X_fr[jfront]
        if presc_bool[gind] == True:
            X_fr[ifront] = presc_val[gind]
            R[gind] = - resid + pivot * presc_val[gind]
        else:
            X_fr[ifront] = resid / pivot
        X[gind] = X_fr[ifront]
    return X, R



# -*- coding: utf-8 -*-
import numpy as np
from numpy import concatenate as join
import awkward as ak
from awkward import unflatten as build
import numpy as np

from .utils import count_cols

def shape(arr): return arr.shape[:2]
def cut(shp): return np.full(shp[0], shp[1], dtype=np.int64)
def flatten(arr): return arr.flatten()


class JaggedArray(ak.Array):
    """
    A numba-jittable class that handles matrices with a variable 
    number of columns per row.
    
    The class is actually an interface to `awkward.Array`,
    with some additional features, specific to 2d jagged arrays.
    
    Parameters
    ----------
    data : Iterable
        Some jagged data.
    
    behaviour
        Passed foreard to `awkward.Array`, see it's documentation
        for the details.
        
    with_name
        Passed foreard to `awkward.Array`, see it's documentation
        for the details.
        
    check_valid
        Passed foreard to `awkward.Array`, see it's documentation
        for the details.
        
    cache
        Passed foreard to `awkward.Array`, see it's documentation
        for the details.
        
    kernels
        Passed foreard to `awkward.Array`, see it's documentation
        for the details.
        
    cuts : Iterable
        An iterable that tells how to unflatten an 1d array into a 
        2d jagged shape.
        
    Returns
    -------
    JaggedArray
                
    Examples
    --------
    >>> from dewloosh.math.sparse import JaggedArray
    
    """

    def __init__(self, data, behavior=None, with_name=None,
                 check_valid=False, cache=None, kernels=None,
                 cuts=None) -> 'JaggedArray':
        if isinstance(data, np.ndarray):
            nD = len(data.shape)
            assert nD <= 2, "Only 2 dimensional arrays are supported!"
            if nD == 1:
                assert isinstance(cuts, np.ndarray)
                data = build(data, cuts)
        elif isinstance(data, list):
            assert all(map(lambda arr: len(arr.shape) == 2, data)), \
                "Only 2 dimensional arrays are supported!"
            # NOTE - implementaion 1
            # > Through the back door, but this is probably the cleanest solution of all.
            # > It only requires to create one python list, without further operations on it.
            # NOTE This line is one of the most unreadable things I've ever done.
            data = build(join(list(map(flatten, data))), join(list(map(cut, map(shape, data)))))
            # NOTE - implementaion 2
            #from operator import add
            #from functools import reduce
            # > This also works, but requires data to jump back and forth just to
            # > have a merged list of lists. It also requires to add nested python lists,
            # > which is probably not the quickest operation in the computational world.
            #data = ak.from_iter(reduce(add, map(lambda arr : ak.Array(arr).to_list(), data)))
            # NOTE - implementaion 3
            # > This also works at creation, but fails later at some operation due to
            # > the specific layout generated by ak.concatenate
            #data = ak.concatenate(list(map(lambda arr : ak.Array(arr), data)))
        super().__init__(data, behavior=behavior, with_name=with_name,
                         check_valid=check_valid, cache=cache, kernels=kernels)

    def is_jagged(self):
        widths = self.widths()
        return not np.all(widths == widths[0])

    def widths(self):
        assert self.ndim == 2, "Only 2 dimensional arrays are supported!"
        return count_cols(self)

    def flatten(self, return_cuts=False):
        """
        Returns the flattened equivalent of the array.
        """
        if return_cuts:
            return self.widths(), ak.flatten(self)
        else:
            return ak.flatten(self)

    def __array_function__(self, func, types, args, kwargs):
        # NOTE Moved to TopologyArray
        # if func == np.unique:
        #    return unique2d(*args, **kwargs)
        return super().__array_function__(func, types, args, kwargs)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        return super().__array_ufunc__(ufunc, method, *inputs, **kwargs)



# -*- coding: utf-8 -*-
from typing import Callable
import numpy as np

from dewloosh.math.hist import *
from dewloosh.math.utils import *

__version__ = "0.0.1b"
__description__ = "A math library for numerical and symboliccalculations."


def squeeze_if_array(arr): return np.squeeze(
    arr) if isinstance(arr, np.ndarray) else arr


def squeeze(default=True):
    def decorator(fnc: Callable):
        def inner(*args, **kwargs):
            if kwargs.get('squeeze', default):
                res = fnc(*args, **kwargs)
                if isinstance(res, tuple):
                    return list(map(squeeze_if_array, res))
                return squeeze_if_array(res)
            else:
                return fnc(*args, **kwargs)
        return inner
    return decorator


def config(*args, **kwargs):
    def decorator(fnc: Callable):
        def inner(*args, **kwargs):
            return fnc(*args, **kwargs)
        return inner
    return decorator


def is_none_or_false(a):
    if isinstance(a, bool):
        return not a
    elif a is None:
        return True
    return False
